/*
  fast cd program v1.1 - 4/12/2002

  wadeb@wadeb.com
  www.wadeb.com

  the parameter passed is one of the following:
  - an absolute path
  - a relative path

 4/12/2002 v1.1 - some bugs fixed, relative paths work now.
 3/1/2002  v1.0 - initial release
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>

#define DELIM_CHAR '/'
#define DELIM_STR "/"

#if 1 
void wtb_printf( char* fmt, ... ) {}
#else
#define wtb_printf printf
#endif

/* buffer for building up the path tree and storing the argument.  if the buffer exceeds, it
   will cleanly fail that branch but will continue to try shorter paths. */
#define BUFSIZE 4096
char path_work[BUFSIZE + FILENAME_MAX * 2];
char step_work[BUFSIZE + FILENAME_MAX * 2];
char temp[BUFSIZE + FILENAME_MAX * 2];
char temp2[BUFSIZE + FILENAME_MAX * 2];
char search_work[BUFSIZE];

#define FOUND 1
#define NOTFOUND 0

int isdir( char* path, u_int ofs, char* name )
{
  struct stat entry_stat;
  strcpy( temp, path );
  strcpy( temp + ofs, name );
  wtb_printf( "isdir: %s\n", temp );
  if ( stat( temp, &entry_stat ) == 0 )
    return entry_stat.st_mode & S_IFDIR;
  else
    return 0;
}

/* enhanced version that substitutes "./" in case of an empty string. */
void wtb_opendir( char* path )
{
}

int match_pattern( char* string, char* pattern )
{
  char c, p, l;

  for ( ;; ) 
  {
    switch (p = *pattern++) 
    {
    case 0:                             
      return *string ? 0 : 1;
      
    case '*':
      while ( *string )                     
      {               
        if ( match_pattern( string++, pattern ) )
          return 1;
      }
      return match_pattern( string, pattern );
      
    case '?':
      if ( *string++ == 0 )
        return 0;         
      break;
      
    case '[':
      if ( ( c = *string++ ) == 0 )
        return 0;                 
      
      c = toupper( c );
      l = 0;
      while ( p = *pattern++ ) 
      {
        if ( p == ']' )          
          return 0;             
        
        if ( p == '-' )        
        {             
          p = *pattern;       
          if ( p == 0  ||  p == ']' )
            return 0;        
          
          if ( c >= l  &&  c <= p )
            break;          
        }
        
        l = p;
        if ( c == p )      
          break;          
      }
      
      while ( p  &&  p != ']' )
        p = *pattern++;       
      break;
      
    default:
      c = *string++;
      if ( toupper( c ) != toupper( p ) ) 
        return 0;                   
      break;
    }
  }
}

/* path_end is the current position in the 'path tree' including the root directory.
   step is the current location in the wildcard string. */
int resolve_path( char* path_end, char* step )
{
  DIR* dir;
  struct dirent* entry;
  char* next_step = NULL;
  
  wtb_printf( "resolve_path: path_work:%s step:%s\n", path_work, step );

  /* string buffer guards. */
  if ( path_end - path_work > BUFSIZE )
    return NOTFOUND;
  if ( step - step_work > BUFSIZE )
    return NOTFOUND;

  /* delimit the current step in the path. */
  next_step = strchr( step, DELIM_CHAR );
  if ( next_step )
  {
    *next_step = '\0';
    next_step++;
  }

  if ( strlen( step ) == 0 )
    return FOUND;
    
  /* first try to open the step work explicitly - works for '..' as well as
     giving preference to exact matches */
  if ( isdir( path_work, path_end - path_work, step ) )
  {
    strcpy( path_end, step );
    strcat( path_end, DELIM_STR );
    
    wtb_printf( "match: %s\n", path_work );
    
    /* positive match.  if it's the last step, return all the way up. */
    if ( !next_step || !strlen( next_step ) )
      return FOUND;

    /* otherwise try and resolve the next step. */
    if ( resolve_path( path_end + strlen( path_end ), next_step ) == FOUND )
      return FOUND;

    *(path_end) = '\0';  
    wtb_printf( "trying another entry from %s: %s\n", path_work, entry->d_name );
  }
    
  /* if that possibility failed, find all subdirectory entries in this path */
  wtb_printf( "exact match failed, pattern matching.\n" );
  
  dir = opendir( path_work );
  wtb_printf( "opendir returned %x\n", dir );
  if ( !dir )
    goto exit;

  /* use the path name as a scratch area to append the * */
  strcpy( path_end, step );
  strcat( path_end, "*" );
  
  /* scan through all subdirs and try to match the wildcard */
  entry = readdir( dir );
  while ( entry )
  {
    for ( ;; )
    {
      if ( !entry )
        break;
      wtb_printf( "entry: %s against %s\n", entry->d_name, path_end );
      if ( match_pattern( entry->d_name, path_end ) )
      {
        if ( isdir( path_work, path_end - path_work, entry->d_name ) )
          break;
        wtb_printf( "after match: %s\n", path_work );
      }
      entry = readdir( dir );
    }
    if ( !entry )
      goto exit;

    strcpy( path_end, entry->d_name );
    strcat( path_end, DELIM_STR );
    
    wtb_printf( "match: %s\n", path_work );
    
    /* positive match.  if it's the last step, return all the way up. */
    if ( !next_step || !strlen( next_step ) )
      return FOUND;

    /* otherwise try and resolve the next step. */
    if ( resolve_path( path_end + strlen( path_end ), next_step ) == FOUND )
      return FOUND;

    strcpy( path_end, step );
    strcat( path_end, "*" );
  
    entry = readdir( dir );
    wtb_printf( "trying another entry from %s: %s\n", path_work, entry->d_name );
  }

exit:
  if ( dir )
    closedir( dir );

  /* restore the backslash and try another branch. */
  if ( next_step )
    *(next_step - 1) = DELIM_CHAR;

  return NOTFOUND;
}

/* changes directory and exits */
void change_dir( char* dir )
{
  printf( "%s\n", dir );
  exit(0);
}

int main( int argc, char* argv[] )
{
  char* search;

  if ( argc < 2 )
    exit(1);

  strncpy( step_work, argv[1], BUFSIZE );
  step_work[BUFSIZE - 1] = '\0';

  /* see if they requested the root dir */
  if ( step_work[0] == DELIM_CHAR )
  {
    strcpy( path_work, DELIM_STR );
	if ( resolve_path( path_work + strlen( path_work ), step_work + 1 ) == FOUND )
	  change_dir( path_work );
	else
	  return 0;
  }

  /* see if they requested a relative dir */
  if ( step_work[0] == '.' )
  {
    strcpy( path_work, "" );
	if ( resolve_path( path_work + strlen( path_work ), step_work ) == FOUND )
	  change_dir( path_work );
	else
	  return 0;
  }

  /* try relative to the current directory */
  strcpy( path_work, "./" );
  if ( resolve_path( path_work + strlen( path_work ), step_work ) == FOUND )
    change_dir( path_work );

  wtb_printf( "trying search environment var\n" );
  
  /* then try using the SEARCH environment var. */
  search = getenv( "search" );
  if ( !search )
    search = getenv( "SEARCH" );
  if ( !search )
    change_dir( argv[1] );

  wtb_printf( "search environment var: %s\n", search );
  
  strncpy( search_work, search, BUFSIZE );
  search_work[BUFSIZE - 1] = '\0';

  search = search_work;
  while ( search && strlen( search ) )
  {
    char* next_search;
    char* search_end;

    /* trim leading spaces. */
    while ( ' ' == *search )   
      search++;

    /* delimit the entry */
    next_search = strchr( search, ':' );
    if ( !next_search )
      next_search = strchr( search, ';' );
      
    if ( next_search )
    {
      *next_search = '\0';
      next_search++;
    }

    /* trim trailing spaces and slash. */
    search_end = search + strlen( search ) - 1;
    while ( ' ' == *search_end || DELIM_CHAR == *search_end )  
      *(search_end--) = '\0';

    wtb_printf( "trying search path %s\n", search );
    
    strcpy( path_work, search );
    strcat( path_work, DELIM_STR );
    if ( resolve_path( path_work + strlen( path_work ), step_work ) == FOUND )
      change_dir( path_work );

    search = next_search;
  }

  change_dir( argv[1] );
  return 0;
}                          
